/*
 firmware.h - Device Firmware Update Library

 Copyright (c) 2021 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#ifndef _FIRMWARE_H_
#define _FIRMWARE_H_

#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <error.h>
#include "serial.h"

//
// Firmware Address Range
//
#define FIRMWARE_ADDR_START 0x0400
#define FIRMWARE_ADDR_END   0x4000
//
// Bootloader Row Size
//
#define FIRMWARE_SIZE_ROW 64
//
// Bootloader Commands
//
#define FIRMWARE_CMD_ERASE 0x45
#define FIRMWARE_CMD_RESET 0x52
//
// Bootloader Status codes
//
#define FIRMWARE_STATUS_OK               1
#define FIRMWARE_STATUS_INVALID_COMMAND  2
#define FIRMWARE_STATUS_INVALID_CHECKSUM 3
#define FIRMWARE_STATUS_VERIFY_FAILED_L  4
#define FIRMWARE_STATUS_VERIFY_FAILED_H  5

class Firmware : public Serial
{
  private:

    ssize_t check(ssize_t status)
    {
      if (status < 0)
        error(strerror(status));
      return status;
    }

    // Hide method
    ssize_t
    read(void *buf, size_t len) override
    {
      ssize_t rv = check(Serial::read(buf, len));
      if ((rv >= 0) && (rv != (ssize_t)len))
        error("READ-TIMEOUT\n");
      return rv;
    }

    // Hide method
    ssize_t
    write(const void *buf, size_t len) override
    {
      ssize_t rv = check(Serial::write(buf, len));
      if ((rv >= 0) && (rv != (ssize_t)len))
        error("WRITE-IOERROR\n");
      return rv;
    }

    bool
    status(void)
    {
      uint8_t status;
      if (read(&status, sizeof(status)) == (ssize_t)sizeof(status))
      {
        switch (status)
        {
          case FIRMWARE_STATUS_OK:
            return true;
          case FIRMWARE_STATUS_INVALID_COMMAND:
            error("INVALID_COMMAND\n");
            break;
          case FIRMWARE_STATUS_INVALID_CHECKSUM:
            error("INVALID_CHECKSUM\n");
            break;
          case FIRMWARE_STATUS_VERIFY_FAILED_L:
            error("VERIFY_FAILED_L\n");
            break;
          case FIRMWARE_STATUS_VERIFY_FAILED_H:
            error("VERIFY_FAILED_H\n");
            break;
          default:
            error("UNKOWN ERROR\n");
            break;
        }
      }
      return false;
    }

    bool
    invoke(const uint8_t *buf, size_t len)
    {
      return write(buf, len) == (ssize_t)len ? status() : false;
    }

    bool
    program(const uint8_t *code, uint32_t addr, uint32_t size)
    {
      struct __attribute__((packed))
      {
        uint8_t low;
        uint8_t high;
        uint8_t csum;
        uint8_t mode;
      } request = {(uint8_t)(addr >> 1), (uint8_t)(addr >> 9), 0, FIRMWARE_CMD_ERASE};
      if (code)
      {
        // calculate checksum
        for (uint32_t i = 0; i < size; ++i)
          request.csum -= code[i];
      }
      // set write params
      if (!invoke((uint8_t *)&request, sizeof(request)))
        return false;
      // write code
      return code ? invoke(code, size) : true;
    }

  public:

    Firmware(void)
    {
    }

    virtual
    ~Firmware(void)
    {
    }

    bool
    reset(void)
    {
      uint8_t cmd = FIRMWARE_CMD_RESET;
      return write(&cmd, sizeof(cmd)) == (ssize_t)sizeof(cmd);
    }

    bool
    erase(void)
    {
      return program(NULL, FIRMWARE_ADDR_START, FIRMWARE_SIZE_ROW);
    }

    bool
    upload(const uint8_t *code, uint32_t start, uint32_t end)
    {
      if ((start < FIRMWARE_ADDR_START) || (end > FIRMWARE_ADDR_END))
      {
        error("invalid code address range\n");
        return false;
      }
      start &= ~(FIRMWARE_SIZE_ROW - 1);
      printf("upload address range (0x%04X - 0x%04X)\n", start, end - 1);
      while (start < end)
      {
        printf(".");
        if (!program(code + start, start, FIRMWARE_SIZE_ROW))
        {
          printf("\n");
          error("stopped address is 0x%04X\n", start);
          return false;
        }
        start += FIRMWARE_SIZE_ROW;
      }
      printf("\nupload success and reset\n");
      return reset();
    }

    void error(const char *format,...)
    {
      va_list ap;
      va_start(ap, format);
      fprintf(stderr, "[ERROR] ");
      vfprintf(stderr, format, ap);
      va_end(ap);
    }
};

#endif /* _FIRMWARE_H_ */
