/*
 i16hex.h - Intel 16bit Hex File Library

 Copyright (c) 2021 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#ifndef _IHEX16_H_
#define _IHEX16_H_

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>

#ifndef __linux__
  #if SIZE_MAX == UINT64_MAX
    #define ssize_t long long
  #else
    #define ssize_t int
  #endif
#endif

enum
{
  I16HEX_TYPE_DATA = 0,
  I16HEX_TYPE_END  = 1,
  I16HEX_TYPE_IH32 = 4,
};

class I16Hex
{
  private:

    uint32_t _start;
    uint32_t _end;
    uint8_t  _bytes[65536];

    template<typename T>
    const uint8_t *toxdigit(const uint8_t *p, T &val)
    {
      val = 0;
      for (uint8_t i = 0; i < sizeof(T) * 2; ++i)
      {
        uint8_t c = *p++;
        if (!isxdigit(c))
          return NULL;
        val = (val << 4) | (c <= '9' ?  c - '0' : (c & ~0x20) - ('A' - 10));
      }
      return p;
    }

  public:

    I16Hex(void)
    {
      clear();
    }

    virtual
    ~I16Hex(void)
    {
    }

    void
    clear(void)
    {
      _start = _end = 0;
      memset(_bytes, 0xFF, sizeof(_bytes));
    }

    bool
    load(const char *path)
    {
      bool eor = false;
      FILE *fp = fopen(path, "r");
      if (fp)
      {
        uint32_t base = 0;
        uint16_t addr;
        uint8_t  nlen, type, csum, sum, cnt;
        uint8_t  line[(sizeof(nlen) + sizeof(addr) + sizeof(type) + UINT8_MAX + sizeof(csum)) * 2 + 4];
        clear();
        while (!eor && fgets((char *)line, sizeof(line), fp))
        {
          const uint8_t *p = line;
          if (*p++ == ':')
          {
            if ((p = toxdigit<__typeof__(nlen)>(p, nlen)) == NULL)
              break; // invalid format.
            if ((p = toxdigit<__typeof__(addr)>(p, addr)) == NULL)
              break; // invalid format.
            if ((p = toxdigit<__typeof__(type)>(p, type)) == NULL)
              break; // invalid format.
            for (cnt = sum = 0; cnt < nlen; ++cnt)
            {
              __typeof__(line[0]) b;
              if ((p = toxdigit<__typeof__(b)>(p, b)) == NULL)
                break; // invalid format.
              sum += line[cnt] = b;
            }
            if (cnt < nlen)
              break; // invalid format.
            if ((p = toxdigit<__typeof__(csum)>(p, csum)) == NULL)
              break; // invalid format.
            if ((uint8_t)(nlen + (addr >> 8) + (addr >> 0) + type + csum + sum))
              break; // invalid check sum.
            switch (type)
            {
              case I16HEX_TYPE_DATA:
                if (base + addr + nlen <= sizeof(_bytes))
                  memcpy(_bytes + addr, line, nlen);
                break;
              case I16HEX_TYPE_END:
                for (_end = sizeof(_bytes); (_end > 0) && (_bytes[_end - 1] == 0xFF); --_end)
                  continue;
                for (_start = 0; (_end > _start) && (_bytes[_start] == 0xFF); ++_start)
                  continue;
                eor = true;
                break;
              case I16HEX_TYPE_IH32:
                base = ((uint32_t)line[0] << 24) | ((uint32_t)line[1] << 16);
                break;
              default:
                break; // ignored.
            }
          }
        }
        fclose(fp);
      }
      return eor;
    }

    uint32_t
    start(void)
    {
      return _start;
    }

    uint32_t
    end(void)
    {
      return _end;
    }

    const uint8_t *
    bytes(void)
    {
      return _bytes;
    }
};

#endif /* _IHEX16_H_ */
